#!/usr/bin/python

'''
Program for calculation of Cij components of elastic 
tensor from the strain-stress relation.
 
This code is (C) 2002-2006 by Pawel T. Jochym
and it is licensed to use and redistribute under 
the terms of the GNU General Public License v2
See http://www.gnu.org/licenses/licenses.html 
for the details of the terms and conditions.

The strain components here are ordered in non-standard way
this is for historical reasons and does not metter realy
The ordering is: uxx, uyy, uzz, uxy, uxz, uyz  
The same goes for the ordering of Cij components
It is:

    C   C   C   C   C   C   C   C   C   C   C   C   C
     11  22  33  12  13  23  44  55  66  16  26  36  45

These functions define the symmetry of the Cij matrix.
The matrix is N columns by 6 rows where the columns
corespond to independent elastic constants of the given
crystal, while the rows corespond to the canonical deformations
of a crystal. The elements are the second partial derivatives of 
the free energy formula for the crystal written down as a 
quadratic form of the deformations with respect to elastic
constant and deformation. 

IMPORTANT!

The elements for deformations uxy, uxz, uyz have to be divided
by 2 to properly match the usual definition of elastic constants.

See: L.D. Landau, E.M. Lifszyc, "Theory of elasticity"

There is some usefull summary also at: 
http://scienceworld.wolfram.com/physics/Elasticity.html
'''

import Numeric
import string
import sys
import LinearAlgebra

from Numeric import array, matrixmultiply, transpose, reshape
from LinearAlgebra import inverse, linear_least_squares


def regular(u):
    '''
    Regular (a.k.a. cubic) lattice
    '''
    
    uxx, uyy, uzz, uxy, uxz, uyz = u[0],u[1],u[2],u[3],u[4],u[5]
    return array(
    [[uxx,     uyy + uzz,     0], 
     [uyy,     uxx + uzz,     0], 
     [uzz,     uxx + uyy,     0], 
     [0,         0,         2*uxy], 
     [0,         0,         2*uxz], 
     [0,         0,         2*uyz]])

def tetragonal(u):
    uxx, uyy, uzz, uxy, uxz, uyz = u[0],u[1],u[2],u[3],u[4],u[5]
    return array(
    [[uxx,   0,    uyy,  uzz,      0,      0],
     [uyy,   0,    uxx,  uzz,      0,      0],
     [0,     uzz,  0,    uxx+uyy,  0,      0],
     [0,     0,    0,    0,        0,      2*uxy],
     [0,     0,    0,    0,        2*uxz,  0],
     [0,     0,    0,    0,        2*uyz,  0]])
 

def orthorombic(u):
    uxx, uyy, uzz, uxy, uxz, uyz = u[0],u[1],u[2],u[3],u[4],u[5]
    return array(
    [[uxx,    0,    0,  uyy,  uzz,    0,    0,    0,    0],
    [0,     uyy,    0,  uxx,    0,  uzz,    0,    0,    0],
    [0,       0,  uzz,    0,  uxx,  uyy,    0,    0,    0],
    [0,       0,    0,    0,    0,    0,    0,    0,2*uxy],
    [0,       0,    0,    0,    0,    0,    0,2*uxz,    0],
    [0,       0,    0,    0,    0,    0,2*uyz,    0,    0]])
 

def rhombohedral(u):
    '''
    The matrix is constructed based on the approach from L&L
    using xi=x+iy ; eta=x-iy auxiliary coordinates. 
    There is some doubt about the C14 constants. 
    This is still to be verified at a later stage.
    The order of constants is as follows:
    C   C   C   C   C   C  
     11  33  12  13  44  14
    '''
    
    uxx, uyy, uzz, uxy, uxz, uyz = u[0],u[1],u[2],u[3],u[4],u[5]
    return array(
    [[   uxx,   0,    uyy,     uzz,     0,   2*uxz],
     [   uyy,   0,    uxx,     uzz,     0,  -2*uxz],
     [     0, uzz,      0, uxx+uyy,     0,   0],
     [ 2*uxy,   0, -2*uxy,       0,     0,  -4*uyz],
     [     0,   0,      0,       0, 2*uxz,   2*(uxx-uyy)],
     [     0,   0,      0,       0, 2*uyz,  -4*uxy]])

def hexagonal(u):
    '''
    The matrix is constructed based on the approach from L&L
    using xi=x+iy ; eta=x-iy auxiliary coordinates. 
    This is still to be verified at a later stage.
    The order of constants is as follows:
    C   C   C   C   C
     11  33  12  13  44
    '''
    
    uxx, uyy, uzz, uxy, uxz, uyz = u[0],u[1],u[2],u[3],u[4],u[5]
    return array(
    [[   uxx,   0,    uyy,     uzz,     0   ],
     [   uyy,   0,    uxx,     uzz,     0   ],
     [     0, uzz,      0, uxx+uyy,     0   ],
     [ 2*uxy,   0, -2*uxy,       0,     0   ],
     [     0,   0,      0,       0, 2*uxz   ],
     [     0,   0,      0,       0, 2*uyz   ]])

def monoclinic(u):
    '''
    Monoclinic group, the order is:
    C   C   C   C   C   C   C   C   C   C   C   C   C
     11  22  33  12  13  23  44  55  66  16  26  36  45
    '''
    
    uxx, uyy, uzz, uxy, uxz, uyz = u[0],u[1],u[2],u[3],u[4],u[5]
    return array(
    [[uxx,  0,  0,uyy,uzz,  0,    0,    0,    0,uxy,  0,  0,  0],
     [  0,uyy,  0,uxx,  0,uzz,    0,    0,    0,  0,uxy,  0,  0],
     [  0,  0,uzz,  0,uxx,uyy,    0,    0,    0,  0,  0,uxy,  0],
     [  0,  0,  0,  0,  0,  0,2*uyz,    0,    0,  0,  0,  0,uxz],
     [  0,  0,  0,  0,  0,  0,    0,2*uxz,    0,  0,  0,  0,uyz],
     [  0,  0,  0,  0,  0,  0,    0,    0,2*uxy,uxx,uyy,uzz,  0]])



def ReadFile( fn ):
    '''
    Read data from the file with data extracted from the calculation
    
    The format is the same as for the mathematica module:
    A list of lists containing data for each pressure point
    and each deformation, preceded by the reference geometry data
    in the same format as the deformation data. The structure is as 
    follows (unfortunately it uses {} as list delimiters and 
    must be translated before evaluating):
    '{n,            # number of pressure points
     {            # List of pressure points
      { p, e, v,        # pressure, energy, cell volume
       {{Ax, Ay, Az},    # Lattice matrix of the reference geometry
        {Bx, By, Bz},    # for pressure P.
        {Cx, Cy, Cz}},
       {sxx, syy, szz,     # Stress tensor of the reference config
        syz, sxz, sxy}},
      { ... {},{}}'        # Data for the first deformation
                  # The geometry is a deformation (!)
                # not lattice vectors.
    The value returned by the function is a resp. tuple of tuples
    '''
    
    f=open(fn+".math.data", "r")
    fd=string.translate(f.read(),string.maketrans("{}","()"))
    return eval(fd)


def WriteOutResult( cij, fn ):
    name=fn+".data.txt"
    f=open(name,"w+")
#    print cij[0][3]
    for k in range(0,len(cij)):
        f.write("---- P:"+str(cij[k][0])+" E:"+str(cij[k][1])+" V:"+str(cij[k][2])+"\n")
        f.write("# Cij, Sij, Bij, Sing.Val. (Ordering: 11, 22, 33, 12, 13, 23, 44, 55, 66, 16, 26, 36, 45) \n") 
#        print cij[k][3]
        for i in range(0,cij[k][3][4]):
            f.write(str(cij[k][3][0][i])+"  "+
                str(cij[k][3][1][i])+"  "+
                str(cij[k][3][2][i])+"  "+
                str(cij[k][3][5][i])+"\n")
#    print cij
    f.close()
    name=fn+".Cij.data"
    f=open(name,"w+")
#    print cij[0][3]
    for k in range(0,len(cij)):
        f.write(str(cij[k][0])+" "+str(cij[k][1])+" "+str(cij[k][2]))
        for i in range(0,cij[k][3][4]):
            f.write(" "+str(cij[k][3][0][i]))
        f.write("\n");
#    print cij
    f.close()
    name=fn+".Sij.data"
    f=open(name,"w+")
#    print cij[0][3]
    for k in range(0,len(cij)):
        f.write(str(cij[k][0])+" "+str(cij[k][1])+" "+str(cij[k][2]))
        for i in range(0,cij[k][3][4]):
            f.write(" "+str(cij[k][3][1][i]))
        f.write("\n");
#    print cij
    f.close()
    name=fn+".Bij.data"
    f=open(name,"w+")
#    print cij[0][3]
    for k in range(0,len(cij)):
        f.write(str(cij[k][0])+" "+str(cij[k][1])+" "+str(cij[k][2]))
        for i in range(0,cij[k][3][4]):
            f.write(" "+str(cij[k][3][2][i]))
        f.write("\n");
#    print cij
    f.close()
    return


def CalcCij( dat, symet=regular ):
    '''
    Calculate Cij from data and structure symmetry
    '''

    clst=[]
    for i in range(1,dat[0]+1):            # Go over all pressure points
        p, e, v=dat[i][0][0:3]            # Pressure, energy, volume
        m=inverse(array(dat[i][0][3]))        # Inverse of the lattice matrix for the pressure P
        s=array(dat[i][0][4])            # Stress "vector" for the reference geometry at pressure P
#        print p, e, v,"\n", m,"\n", s
        ul=[]
        sl=[]
        for k in range(1,len(dat[i])):        # Go over all deformations for pressure P
            d=dat[i][k][3]            # Deformation matrix
            t=matrixmultiply(m,d)        # Transform to relative changes
            t=-(t+transpose(t))/2        # Symetrize
#            print t
            ul.append(array([t[0][0],t[1][1],t[2][2],  # Rewrite in the "6-vector" notation
                t[0][1],t[0][2],t[1][2]]))       # Note the different ordering of the 
                                   # components. This to match the 
                                   # definitions "symet" functions above
                                   # this only maters for "symetrization"
                                   # so I've left it here from the previous
                                   # version of the code to not mess
                                   # with something that works
                                   
#            print dat[i][k][4]
            sl.append(array(dat[i][k][4])-s)
#            sl.append(array(dat[i][k][4])-array([p,p,p,0,0,0]))    # Remove the hydrostatic component 
                                        # from the stress tensor
        eqm=array(map(symet,ul))        # Build the equation matrix, mapping symetries over
                            # the deformations
#        print eqm[0].shape, eqm.shape
        eqm=reshape(eqm,(eqm.shape[0]*eqm.shape[1],eqm.shape[2]))  # Make it a proper linear equation matrix
#        print eqm
        slm=reshape(array(sl),(-1,))                   # The same for the RHS
#        print eqm.shape, slm.shape
#        print slm
        Bij = linear_least_squares(eqm,slm)               # Solve the equation with SVD
#        print Bij
        
        # Calculate elastic constants from Birch coeff.
        if (symet == orthorombic):
            Cij = Bij[0] - array([-p,-p,-p, p, p, p,-p,-p,-p])
        elif (symet == tetragonal):
            Cij = Bij[0] - array([-p,-p, p, p,-p,-p])
        elif (symet == regular):
            Cij = Bij[0] - array([-p, p,-p])
        elif (symet == rhombohedral):
            Cij = Bij[0] - array([-p,-p,p,p,-p,p])
        elif (symet == hexagonal):
            Cij = Bij[0] - array([-p,-p,p,p,-p])
                
#        print Cij
        # Quick hack to print Sij 
        if (symet == tetragonal):
            c11, c33, c12, c13, c44, c66=Cij
            #print c11, c33, c12, c13, c44, c66
            Sij=inverse(array(
                   [[c11,c12,c13,  0,  0,  0],
                    [c12,c11,c13,  0,  0,  0],
                    [c13,c13,c33,  0,  0,  0],
                    [  0,  0,  0,c44,  0,  0],
                    [  0,  0,  0,  0,c44,  0],
                    [  0,  0,  0,  0,  0,c66]]))
        elif (symet == orthorombic):
            c11, c22, c33, c12, c13, c23, c44, c55, c66=Cij
            #print c11, c33, c12, c13, c44, c66
            Sij=inverse(array( 
                   [[c11,c12,c13,  0,  0,  0],
                    [c12,c22,c23,  0,  0,  0],
                    [c13,c23,c33,  0,  0,  0],
                    [  0,  0,  0,c44,  0,  0],
                    [  0,  0,  0,  0,c55,  0],
                    [  0,  0,  0,  0,  0,c66]]))
        elif (symet == regular):
            c11, c12, c44 =Cij
            #print c11, c12, c44
            Sij=inverse(array( 
                   [[c11,c12,c12,  0,  0,  0],
                    [c12,c11,c12,  0,  0,  0],
                    [c12,c12,c11,  0,  0,  0],
                    [  0,  0,  0,c44,  0,  0],
                    [  0,  0,  0,  0,c44,  0],
                    [  0,  0,  0,  0,  0,c44]]))
        elif (symet == rhombohedral):
            # I am not sure if the matrix is correct
            # TODO: check the matrix elements in lower part (>3)
            c11, c33, c12, c13, c44, c14=Cij
            #print c11, c33, c12, c13, c44, c14
            Sij=inverse(array( 
                   [[c11, c12,c13, c14,  0,  0],
                    [c12, c11,c13,-c14,  0,  0],
                    [c13, c13,c33,   0,  0,  0],
                    [c14,-c14,  0, c44,  0,  0],
                    [  0,   0,  0,   0,c44,c14/2],
                    [  0,   0,  0,   0,c14/2,(c11-c12)/2]]))
        elif (symet == hexagonal):
            # I am not sure if the matrix is correct
            # TODO: check the matrix elements in lower part (>3)
            c11, c33, c12, c13, c44=Cij
            #print c11, c33, c12, c13, c44, c14
            Sij=inverse(array( 
                   [[c11, c12,c13,   0,  0,  0],
                    [c12, c11,c13,   0,  0,  0],
                    [c13, c13,c33,   0,  0,  0],
                    [  0,   0,  0, c44,  0,  0],
                    [  0,   0,  0,   0,c44,  0],
                    [  0,   0,  0,   0,  0,(c11-c12)/2]]))
            
            
        # Now convert Sij to the Voight notation
        
        Si=[Sij[0][0],Sij[1][1],Sij[2][2],Sij[0][1],Sij[0][2],Sij[1][2],Sij[3][3],Sij[4][4],Sij[5][5]]
#        print Bij
            
        clst.append((p,e,v,[Cij,Si,Bij[0],Bij[1],Bij[2],Bij[3]]))      # append the calculated data to the result list
    return clst

dat = ReadFile(sys.argv[1])
#regular(uxx, uyy, uzz, uxy, uxz, uyz):
cij=CalcCij( dat, eval(sys.argv[2]) )

WriteOutResult(cij, sys.argv[1])


#print dat[0], dat[1][0][3]
